local titlescene = {}

function titlescene:enter()

  

    --モード表示時の文字フォント。
    font = love.graphics.newFont('PixelMplus10-Bold.ttf',30)
  
    --タイトルロゴの画像、サイズ取得、透明度。
    logo_title = love.graphics.newImage("image/title/PD2title.png")
    logo_width = logo_title:getWidth()
    logo_height = logo_title:getHeight()
    logo_alpha = 0

    --開始時にタイトル用BGMを鳴らす。
    title_BGM = love.audio.newSource("sound/bgm/OLD_VILLAGE.mp3", "stream")
    title_BGM:setLooping(true)
    title_BGM:play()
    title_BGM:setVolume(0.2)

    --ウィンドウのサイズを取得。
    window_width = love.graphics.getWidth()
    window_height = love.graphics.getHeight()

    --タイトルロゴが少しづつ現れるためのタイマー設定。
    logo_feed_timer_plus = 0

    --モード選択項目のアルファ値。
    mode_alpha = 1

    --モード選択時のアイコンのパラメーター。
    cursor = {
        x = 0,
        y = 500,
        image = love.graphics.newImage("image/title/kaeruicon.png"),
        selected_mode = 0,
    }

    --カーソルの移動音。
    cursor_move_SE = love.audio.newSource("sound/se/GB-RPG01-6(Select).mp3","static")
    modeselect_SE = love.audio.newSource("sound/se/GB-Fighting01-04(Attack).mp3", "static")

    --カーソル移動可能かどうかのフラグ。1だと移動不可。
    cursor_canmove = 0

    --モードセレクト可能かどうかのフラグ。モード選択時に1になる。キャラセレクト時に2になる。
    modeselect_flag = 0

    --モード決定時の待ち時間
    decide_waittime = 0
    decide_waittime_MAX = 70

    --選んだキャラを格納する変数。
    select_player = ""


end



function titlescene:update(dt)

    --最初のmain.luaからウインドウサイズ変更用関数の呼び出し。
    windowResize()

    --baton関連動作の更新。
    input:update()

    --デルタタイムをロゴ表示用タイマー変数にプラスする。
    logo_feed_timer_plus = logo_feed_timer_plus + dt
    
    --logo_feed_timer_plusが一定値以上になったらアルファ値を増やして少しづつロゴを表示する。
    if logo_feed_timer_plus >= 1 and  logo_alpha < 1 then
        logo_alpha = logo_alpha + 0.35
        logo_feed_timer_plus = 0
    end

    --ロゴが完全に表示されてから間を置いてモードセレクトの文字が表示される。
    if modeselect_flag == 0 and logo_alpha >= 1 and logo_feed_timer_plus > 2 then
        modeselect_flag = 1
        cursor.selected_mode = 1
    end   




   --上下キーを押してカーソルを動かす。
   -- if modeselect_flag == 1 and input:down('UpButton',0.2,0.2) or 
    --   modeselect_flag == 1 and input:down('DownButton',0.2,0.2) then 
    if modeselect_flag >= 1 and input:pressed "down"  and cursor_canmove == 0 or
        modeselect_flag >= 1 and input:pressed "up" and cursor_canmove == 0  then
        cursor.selected_mode = cursor.selected_mode + 1
        cursor_move_SE:stop()
        cursor_move_SE:play()
    end

    --モード選択のカーソルの位置を設定。
    if cursor.selected_mode >= 1 then
        cursor.x = window_width*(2/5)
    end

    if cursor.selected_mode == 1 then
        cursor.y = window_height*(4/6)
    elseif cursor.selected_mode == 2 then
        cursor.y = window_height*(5/6)
    end

    if cursor.selected_mode >= 3 then
        cursor.selected_mode = 1
    end


    --select_mode=1で決定ボタンを押したらキャラセレクトに移行する処理。
    if cursor_canmove == 0 and cursor.selected_mode >= 1 and input:pressed "mainshot" then
        cursor_canmove = 1
        modeselect_SE:stop()
        modeselect_SE:play()
    end 
    
    --カーソルが動かないとき(モード決定時)は次の項目が出るまで待ち時間を設ける。
    if cursor_canmove == 1 then
        decide_waittime = decide_waittime + 1 
    end

    --NEWGAME選択時は、待ち時間(decide_waittime)が一定値になったらmodeselectフラグを2にしてキャラセレクトを表示する。
    if modeselect_flag == 1 and 
    decide_waittime >= decide_waittime_MAX and 
    cursor.selected_mode == 1 then
            cursor_canmove = 0
            modeselect_flag = 2
            decide_waittime = 0
    end

    --待ち時間の間に項目を点滅させる処理。
    if decide_waittime % 2 == 1 then
        mode_alpha = 0
    else mode_alpha = 1
    end

    --キャラセレクト時に選んだキャラを変数に保存。
    if modeselect_flag == 2 and cursor.selected_mode == 1 then
        select_player = "AMATT"
    elseif modeselect_flag == 2 and cursor.selected_mode == 2 then
         select_player = "HITY"
    end 
   

    if decide_waittime >= decide_waittime_MAX and 
    modeselect_flag == 2 then
        title_BGM:stop()
        --ゲームシーンへ遷移。
         Gamestate.registerEvents()
         set_state "gamescene"
    end
end


function titlescene:draw()

    --simpleScaleの開始。
    simpleScale.set()

    --モードセレクト用の文字を描画。
    if modeselect_flag == 1 then
        love.graphics.setColor(0.8, 1, 1, mode_alpha)
        love.graphics.setFont(font)
        love.graphics.printf("NEW GAME",0, window_height*(4/6), window_width, 'center')
        love.graphics.printf("LOAD GAME",0, window_height*(5/6), window_width, 'center')

        love.graphics.draw(cursor.image, cursor.x,cursor.y,0)
    end    


    if modeselect_flag == 2 then
        love.graphics.setColor(1, 0.8, 1, mode_alpha)
        love.graphics.setFont(font)
        love.graphics.printf("AMATT",0, window_height*(4/6), window_width, 'center')
        love.graphics.printf("HITY",0, window_height*(5/6), window_width, 'center')

        love.graphics.draw(cursor.image, cursor.x,cursor.y,0)
    end    


    --背景色を暗めの青にする。
    love.graphics.setBackgroundColor( 0, 0, 0.4, 0.5 )

    --タイトルロゴの描画。
    love.graphics.setColor(1, 1, 1, logo_alpha)
    love.graphics.draw(logo_title, window_width/2, window_height/4, 0,1,1,logo_width/2,logo_height/2)

    --simpleScaleの終了。
    simpleScale.unSet()
end





return titlescene